@@ -1,6 +1,6 @@
 # Copyright (c) Meta Platforms, Inc. and affiliates
 from collections.abc import Sequence
-from typing import cast
+from typing import cast, Optional
 
 import torch
 from torch.distributed.tensor._dtensor_spec import DTensorSpec
@@ -25,6 +25,7 @@ from torch.distributed.tensor.placement_types import (
     Replicate,
     Shard,
 )
+from torch.utils._typing_utils import not_none
 
 
 aten = torch.ops.aten
@@ -45,15 +46,6 @@ aten = torch.ops.aten
 # ]
 
 
-linear_pointwise_ops = [
-    aten.div.Scalar,  # this op is linear on the first argument, and the second argument is scalar, so it fits as a linear op.
-    aten.div_.Scalar,  # this op is linear on the first argument, and the second argument is scalar, so it fits as a linear op.
-    aten.to.dtype,
-    aten.add.Tensor,
-    aten.add_.Tensor,
-]
-
-
 pointwise_ops = [
     # please keep the entries below alphabetically sorted
     aten.__ilshift__.Scalar,
@@ -142,8 +134,14 @@ pointwise_ops = [
     aten.ceil.out,
     aten.ceil_.default,
     aten.clamp.default,
+    aten.clamp.Tensor,
     aten.clamp.out,
     aten.clamp_.default,
+    aten.clamp_.Tensor,
+    aten.clamp_min.default,
+    aten.clamp_min.Tensor,
+    aten.clamp_max.default,
+    aten.clamp_max.Tensor,
     aten.clip.default,
     aten.clip.out,
     aten.clip_.default,
@@ -297,11 +295,7 @@ pointwise_ops = [
     aten.maximum.out,
     aten.minimum.default,
     aten.minimum.out,
-    aten.mul.Scalar,
-    aten.mul.Tensor,
     aten.mul.out,
-    aten.mul_.Scalar,
-    aten.mul_.Tensor,
     aten.mvlgamma.default,
     aten.mvlgamma.out,
     aten.mvlgamma_.default,
@@ -416,18 +410,36 @@ pointwise_ops = [
     aten.threshold_backward.default,
 ]
 
+# the linear pointwise ops map, key is op, value is the type of linearity
+linear_pointwise_ops = {
+    aten.to.dtype: 0,
+    aten.add.Tensor: 1,
+    aten.add_.Tensor: 1,
+    aten.div.Scalar: 0,
+    aten.div_.Scalar: 0,
+    aten.mul.Scalar: 0,
+    aten.mul_.Scalar: 0,
+    aten.mul.Tensor: 2,
+    aten.mul_.Tensor: 2,
+}
 
-def pointwise_strategy(op_schema: OpSchema, linearity: bool = False) -> OpStrategy:
-    max_shards_strategy_index = -1
+
+def pointwise_strategy(op_schema: OpSchema, linearity: int = -1) -> OpStrategy:
+    followed_strategy_index = -1
     max_shards = -1
     max_ndim = -1
 
     if op_schema.is_inplace_op():
         # inplace op should follow the first arg strategy
         followed_strategy = op_schema.args_schema[0]
+        followed_strategy_index = 0
     elif op_schema.is_out_variant_op():
         # out variant op should follow the out kwarg strategy
         followed_strategy = op_schema.kwargs_schema["out"]
+        # out variant is technically a kwarg for the strategy to follow so it does not
+        # have an "index", we set it to a reasonably large number just to indicate it's
+        # not a valid index
+        followed_strategy_index = 100
     else:
         # normal pointwise op, we choose to follow the arg with
         # the max shards in case operands needs reshard
@@ -442,33 +454,73 @@ def pointwise_strategy(op_schema: OpSchema, linearity: bool = False) -> OpStrate
             if (arg_max_shards > max_shards) or (
                 arg_max_shards == max_shards and arg_max_ndim > max_ndim
             ):
-                max_shards_strategy_index = idx
+                followed_strategy_index = idx
                 max_shards = arg_max_shards
                 max_ndim = arg_max_ndim
 
-        followed_strategy = op_schema.args_schema[max_shards_strategy_index]
+        followed_strategy = op_schema.args_schema[followed_strategy_index]
 
     assert isinstance(followed_strategy, OpStrategy), (
         f"no strategy to follow for {op_schema}!"
     )
     return common_pointwise_strategy(
-        op_schema.args_schema, followed_strategy, linearity
+        op_schema.args_schema,
+        followed_strategy,
+        followed_strategy_index,
+        linearity,
     )
 
 
+def linear_pointwise_strategy(op_schema: OpSchema) -> StrategyType:
+    """
+    Linear pointwise operators can propagate pending reductions.
+    For example, c = add(a, b); if a is pending sum, then c will be
+    pending sum as well without any communication overhead.
+
+    Note that:
+    1. Only unary and binary operations are supported, out variant
+      ops are not supported.
+    2. There're multiple types of linearity, refer to the doc of
+      common_pointwise_strategy for more details.
+    """
+    linearity_type = linear_pointwise_ops.get(op_schema.op, -1)
+    return pointwise_strategy(op_schema, linearity=linearity_type)
+
+
 def common_pointwise_strategy(
     args_schema: Sequence[object],
     followed_strategy: OpStrategy,
-    linearity: bool,
+    followed_strategy_index: int,
+    linearity: int = -1,
+    scalar_tensor_idx: Optional[int] = None,
 ) -> OpStrategy:
+    """
+    Common strategy for pointwise operations.
+
+    Args:
+        args_schema: Input arguments schema
+        followed_strategy: Strategy to follow for output placement
+        followed_strategy_index: Index of the strategy being followed
+        linearity: depending on the operator, we support different types of linearity
+            -1: the operation does not support linearity
+            0: the unary operation that supports linearity, output propagates partial.
+            1: the binary operation supports add linearity, where it requires every operand
+                to be partial, output propagates partial.
+            2: the binary operation supports multiplicative linearity, where it requires
+                the primary operand to be partial, and the other operands to be replicate,
+                output propagates partial.
+        scalar_tensor_idx: Index of the Replicate scalar tensor for which we allow the mesh
+            to be different from the mesh of followed_strategy
+    """
     # handle broadcasting
     common_shape = torch.broadcast_shapes(
         *[arg.shape for arg in args_schema if isinstance(arg, OpStrategy)]
     )
     pointwise_strategy = OpStrategy([])
 
-    for placement_strategy in followed_strategy.strategies:
-        spec_to_follow = placement_strategy.output_spec
+    for op_spec in followed_strategy.strategies:
+        spec_to_follow = op_spec.output_spec
+
         out_placements: list[Placement] = []
         for placement in spec_to_follow.placements:
             if isinstance(placement, Shard):
@@ -476,36 +528,72 @@ def common_pointwise_strategy(
                 common_ndim = len(common_shape)
                 new_shard_dim = common_ndim - len(spec_to_follow.shape) + shard_dim
                 out_placements.append(Shard(new_shard_dim))
-            elif isinstance(placement, Partial) and not linearity:
-                # clear the partial placemnet if op does not support linearity
-                # by default we just replicate the partial, need to see if this
-                # is optimal for all cases
-                out_placements.append(Replicate())
+            elif isinstance(placement, Partial):
+                # note that only partial-sum and partial-avg are supported for linearity
+                partial_supports_linearity = placement.is_partial(
+                    "sum"
+                ) or placement.is_partial("avg")
+                if linearity > 0 and partial_supports_linearity:
+                    # propagate the partial placement
+                    out_placements.append(placement)
+                else:
+                    # clear the partial placement if op does not support linearity
+                    # by default we just replicate the partial, need to see if this
+                    # is optimal for all cases
+                    out_placements.append(Replicate())
             else:
                 out_placements.append(placement)
 
         input_specs: list[DTensorSpec] = []
         redistribute_costs: list[list[float]] = []
-        for arg_idx, input_arg in enumerate(args_schema):
+        for input_idx, input_arg in enumerate(args_schema):
             if isinstance(input_arg, OpStrategy):
+                input_arg_spec = input_arg.strategies[0].output_spec
+
                 # sanity check that all args that follow the same strategy
                 # are on the same DeviceMesh
                 if input_arg.mesh != followed_strategy.mesh:
-                    raise ValueError(
-                        f"Could not run pointwise computation across different mesh: "
-                        f"Found {input_arg.mesh} and {followed_strategy.mesh}!"
-                    )
+                    # For the scalar tensor arg in fused ops, do not follow followed_strategy;
+                    # instead, let the input mesh and the Replicate placements propagate through.
+                    if input_idx == scalar_tensor_idx:
+                        assert all(p == Replicate() for p in input_arg_spec.placements)
+                        input_arg_target_spec = DTensorSpec(
+                            mesh=input_arg.mesh,
+                            placements=input_arg_spec.placements,
+                            tensor_meta=input_arg_spec.tensor_meta,
+                        )
+                        input_specs.append(input_arg_target_spec)
+                        redistribute_costs.append(
+                            generate_redistribute_costs(
+                                input_arg, input_arg_target_spec
+                            )
+                        )
+                        continue
+                    else:
+                        raise ValueError(
+                            f"Could not run pointwise computation across different mesh: "
+                            f"Found {input_arg.mesh} and {followed_strategy.mesh}!"
+                        )
 
                 # every arg follow the out_placements, but need to handle broadcasting
-                input_arg_spec = input_arg.strategies[0].output_spec
                 input_arg_dims_map = infer_broadcast_dims_map(
                     common_shape, input_arg_spec.shape
                 )
+
+                # Determine if this input should convert Partial to Replicate base on linearity
+                should_convert_partial = (
+                    linearity == 2
+                    and input_idx
+                    != followed_strategy_index  # Don't convert the "followed" strategy
+                )
+
                 input_target_placements = map_placements_after_broadcast(
                     tuple(out_placements),
                     common_shape,
                     input_arg_dims_map,
+                    partial_to_replicate=should_convert_partial,
                 )
+
                 input_arg_target_spec = DTensorSpec(
                     mesh=followed_strategy.mesh,
                     placements=input_target_placements,
@@ -529,16 +617,7 @@ def common_pointwise_strategy(
     return pointwise_strategy
 
 
-def linear_pointwise_strategy(op_schema: OpSchema) -> StrategyType:
-    """
-    Linear pointwise operators can propagate pending reductions.
-    For example, c = add(a, b); if a is pending sum, then c will be
-    pending sum as well without any communication overhead.
-    """
-    return pointwise_strategy(op_schema, linearity=True)
-
-
-for op in linear_pointwise_ops:
+for op in linear_pointwise_ops.keys():
     register_op_strategy(op, schema_info=RuntimeSchemaInfo(static_kwargkey=["out"]))(
         linear_pointwise_strategy
     )
@@ -629,15 +708,17 @@ def list_pointwise_strategy(
         OpStrategy: generated strategy
     """
 
-    def args_tuple_strategies(args_schema: tuple[object, ...]) -> list[TupleStrategy]:
+    def args_tuple_strategies(
+        args_schema: tuple[object, ...],
+    ) -> list[Optional[TupleStrategy]]:
         first_arg = args_schema[0]
         assert isinstance(first_arg, TupleStrategy)
-        strategy_len = len(first_arg.childs)
-        tuple_strategies: list[TupleStrategy] = []
+        strategy_len = len(first_arg.children)
+        tuple_strategies: list[Optional[TupleStrategy]] = []
         for arg_idx, arg in enumerate(args_schema):
             if isinstance(arg, TupleStrategy):
                 # every tuple strategy should have the same length
-                assert len(arg.childs) == strategy_len
+                assert len(arg.children) == strategy_len
                 tuple_strategies.append(arg)
             elif isinstance(arg, OpStrategy):
                 if arg_idx > 0:  # implicitly broadcast
@@ -648,19 +729,28 @@ def list_pointwise_strategy(
                     raise RuntimeError(
                         f"list op only supports tuple strategy! {op_schema}"
                     )
+            else:
+                # insert None as placeholder so that the idx of arg is kept
+                tuple_strategies.append(None)
         return tuple_strategies
 
     args_strategies = args_tuple_strategies(op_schema.args_schema)
-    follow_strategy: TupleStrategy = args_strategies[0]
+    follow_strategy: TupleStrategy = not_none(args_strategies[0])
     list_strategy: list[OpStrategy] = []
-    for child_idx, child_strtgy in enumerate(follow_strategy.childs):
+
+    for child_idx, child_strtgy in enumerate(follow_strategy.children):
         assert isinstance(child_strtgy, OpStrategy)
-        args_schema: list[OpStrategy] = [
-            cast(OpStrategy, arg_strategy.childs[child_idx])
+        args_schema: list[Optional[OpStrategy]] = [
+            cast(OpStrategy, arg_strategy.children[child_idx]) if arg_strategy else None
             for arg_strategy in args_strategies
         ]
         pointwise_strategy: OpStrategy = common_pointwise_strategy(
-            args_schema, child_strtgy, linearity
+            args_schema,
+            child_strtgy,
+            linearity,
+            scalar_tensor_idx=_FUSED_OP_SCALAR_IDX
+            if op_schema.op in fused_ops
+            else None,
         )
         list_strategy.append(pointwise_strategy)
     return TupleStrategy(list_strategy)
@@ -694,6 +784,13 @@ fused_ops = [
     aten._fused_adamw_.tensor_lr,
 ]
 
+
+# The state_steps arg of fused adam / adamw is a Replicate scalar tensor, which will be put on
+# the compute_mesh of an op across all parameter groups, even when not all parameter groups
+# are on the same device mesh. This idx will help avoid hitting exceptions or unnecessary
+# redistribute during sharding propagation.
+_FUSED_OP_SCALAR_IDX = 5
+
 for op in fused_ops:
     register_op_strategy(op, schema_info=RuntimeSchemaInfo(needs_pytree=True))(
         list_pointwise_strategy
